DOMAIN_SERVER_URL = 'https://github.com/knative/serving/releases/download/knative-v1.14.0/serving-default-domain.yaml'
KNATIVE_ISTIO_CONTROLLER_URL = 'https://github.com/knative/net-istio/releases/latest/download/release.yaml'
KNATIVE_CRD_URL = 'https://github.com/knative/serving/releases/latest/download/serving-crds.yaml'
KNATIVE_CORE_URL = 'https://github.com/knative/serving/releases/latest/download/serving-core.yaml'

